/*----------------------------------------------------------------------
  File    : xbi.c
  Contents: X11 bridg-it program
  Author  : Christian Borgelt
  History : 2001.06.03 file created from file xskel.c
            2013.08.09 adapted to higher compiler warning level
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Core.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Viewport.h>
#include "menu.h"
#include "dialog.h"
#include "fselect.h"
#include "random.h"
#include "bridgit.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- extensions --- */
#define MINXEXT      115        /* minimal x-extension of window */
#define MINYEXT       83        /* minimal y-extension of window */
#define PTDIST        32        /* distance between support points */
#define PTOFFSET       8        /* offset for support points */
#define PTSIZE         6        /* size of support points */
#define MINSIZE        2        /* minimal size of the board */
#define MAXSIZE       32        /* maximal size of the board */

/* --- error codes --- */
#define E_NONE         0        /* no error */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_FOPEN      (-2)       /* cannot open file */
#define E_FREAD      (-3)       /* read error on file */
#define E_FWRITE     (-4)       /* write error on file */
#define E_INIT       (-5)       /* initialization failed */
#define E_WIDGET     (-6)       /* widget creation failed */
#define E_DIALOG     (-7)       /* cannot create dialog box */
#define E_COLOR      (-8)       /* color allocation failed */
#define E_FONT       (-9)       /* font loading failed */
#define E_GRAPH     (-10)       /* cannot create graphics context */
#define E_UNKNOWN   (-11)       /* unkown error */

#define FMT_MAX      256        /* max. len. of error message format */

/* --- texts --- */
#define BLUEWIN    "Blue wins!" /* messages for notifying */
#define REDWIN     "Red wins!"  /* that the game has ended */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- error message data --- */
  const char *name, *type;      /* name and type of the error message */
  const char *class;            /* resource class of error message */
  const char *dflt;             /* default message */
  int        argc;              /* number of arguments */
} ERRMSG;                       /* (error message) */

/*----------------------------------------------------------------------
  Function Prototypes
----------------------------------------------------------------------*/
static void mouse  (Widget w, XEvent *e, String *s, Cardinal *c);
static void load   (Widget w, XEvent *e, String *s, Cardinal *c);
static void save   (Widget w, XEvent *e, String *s, Cardinal *c);
static void quit   (Widget w, XEvent *e, String *s, Cardinal *c);
static void start  (Widget w, XEvent *e, String *s, Cardinal *c);
static void flip   (Widget w, XEvent *e, String *s, Cardinal *c);
static void undo   (Widget w, XEvent *e, String *s, Cardinal *c);
static void redo   (Widget w, XEvent *e, String *s, Cardinal *c);
static void hint   (Widget w, XEvent *e, String *s, Cardinal *c);
static void setup  (Widget w, XEvent *e, String *s, Cardinal *c);
static void redraw (Widget w, XEvent *e, String *s, Cardinal *c);
static void about  (Widget w, XEvent *e, String *s, Cardinal *c);

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#include "xbi.rsc"              /* fallback resources */

static XtActionsRec actions[] = {
  { "mouse",    mouse    },     /* mouse button pressed */
  { "load",     load     },     /* File    > Load File... */
  { "save",     save     },     /* File    > Save File... */
  { "quit",     quit     },     /* File    > Quit */
  { "redraw",   redraw   },     /* Actions > Redraw */
  { "start",    start    },     /* Actions > New Game */
  { "flip",     flip     },     /* Actions > Flip Sides */
  { "undo",     undo     },     /* Actions > Undo Move */
  { "redo",     redo     },     /* Actions > Redo Move */
  { "hint",     hint     },     /* Actions > Give a Hint */
  { "setup",    setup    },     /* Actions > Setup... */
  { "about",    about    },     /* Help    > About Bridg-It... */
  { "db_close", db_close } };   /* close dialog box */

/* --- error messages --- */
static const ERRMSG errmsgs[] = {   /* error message data */
  /* E_NONE      0 */  { "error",   "none",    "Xbi.Error",
                         "no error",                          0 },
  /* E_NOMEM    -1 */  { "error",   "nomem",   "Xbi.Error",
                         "not enough memory",                 0 },
  /* E_FOPEN    -2 */  { "warning", "fopen",   "Xbi.Warning",
                         "cannot open file:\n%s",             1 },
  /* E_FREAD    -3 */  { "warning", "fread",   "Xbi.Warning",
                         "read error on file:\n%s",           1 },
  /* E_FWRITE   -4 */  { "warning", "fwrite",  "Xbi.Warning",
                         "write error on file:\n%s",          1 },
  /* E_INIT     -5 */  { "error",   "init",    "Xbi.Error",
                         "initialization failed",             0 },
  /* E_WIDGET   -6 */  { "error",   "widget",  "Xbi.Error",
                         "widget creation failed",            0 },
  /* E_DIALOG   -7 */  { "warning", "dialog",  "Xbi.Warning",
                         "cannot create dialog box",          0 },
  /* E_COLOR    -8 */  { "error",   "color",   "Xbi.Error",
                         "color allocation failed",           0 },
  /* E_FONT     -9 */  { "error",   "font",    "Xbi.Error",
                         "font loading failed",               0 },
  /* E_GRAPH   -10 */  { "error",   "gc",      "Xbi.Error",
                         "cannot create graphics context",    0 },
  /* E_UNKNOWN -11 */  { "error", "unknown",   "Xbi.Error",
                         "unknown error",                     0 } };

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static XtAppContext appctx;     /* X11 application context */
static Display *display;        /* display connected to */
static int     screen;          /* screen of the display */
static Window  window;          /* main window */
static Atom    wm_delwin;       /* delete window atom of window mgr. */
static char    fmt[FMT_MAX];    /* format and buffer for messages */
static char    buf[PATH_MAX +FMT_MAX];

/* --- main widgets and menu --- */
static Widget  w_top;           /* top level widget */
static Widget  w_main;          /* main window widget */
static Widget  w_view;          /* viewport widget */
static Widget  w_board;         /* bridg-it board widget */
static MENU    *menu;           /* pull down menu */
static GC      gc;              /* graphics context for maze */

/* --- property sheet --- */
static PSHEET  *psh = NULL;     /* property sheet */

/* --- colors --- */
static unsigned long white;     /* background color */
static unsigned long black;     /* foreground color */
static unsigned long red;       /* color horizontal player */
static unsigned long blue;      /* color vertical   player */
static unsigned long dred;      /* color for panel for horz. player */
static unsigned long dblue;     /* color for panel for vert. player */
static unsigned long green;     /* color for highlighting */

/* --- game variables --- */
static BRIDGIT *bi = NULL;      /* bridg-it game */
static struct {                 /* bridg-it parameters */
  int size;                     /* board size (number of rows/cols.) */
  int first;                    /* player to move first */
  int next;                     /* player to move next */
  int hint;                     /* hint for next move */
} bip = { 5, BI_HORZ, BI_HORZ, BI_NONE };

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static void dcb_nop (Widget widget, XtPointer client, XtPointer call)
{                               /* --- close dialog box (callback) */
  while (XtClass(widget) != transientShellWidgetClass)
    widget = XtParent(widget);  /* find the shell widget */
  XtPopdown(widget);            /* close dialog box */
  mn_enable(menu, MN_MENU, 1);  /* and enable menu bar */
}  /* dcb_nop() */

/*--------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list      va;              /* list of variable arguments */
  const ERRMSG *msg;            /* error message */
  String       argv[8];         /* variable argument vector */
  Cardinal     argc;            /* number of arguments */

  if ((code > 0) || (code < E_UNKNOWN))
    code = E_UNKNOWN;           /* check error code */
  msg = errmsgs -code;          /* get error message data and */
  va_start(va, code);           /* list of variable arguments */
  for (argc = 0; (int)argc < msg->argc; argc++)
    argv[argc] = va_arg(va, String);
  va_end(va);                   /* get message arguments */
  if (strcmp(msg->name, "error") == 0)  /* if name is "error" */
    XtAppErrorMsg(appctx, msg->name, msg->type, msg->class,
      msg->dflt, argv, &argc);  /* print error message */
  else {                        /* if name is "warning" */
    XtAppGetErrorDatabaseText(appctx, msg->name, msg->type,
      msg->class, msg->dflt, fmt, FMT_MAX, NULL);
    va_start(va, code);         /* list of variable arguments */
    vsprintf(buf, fmt, va);     /* get and format message */
    va_end(va);                 /* end variable argument evaluation */
    mn_enable(menu, MN_MENU,0); /* disable menu bar and show dialog */
    if (db_alert(w_top, dcb_nop, buf) != 0) {
      mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
      XtAppWarningMsg(appctx, msg->name, msg->type, msg->class,
        msg->dflt, argv, &argc);/* if the alert dialog box */
    }                           /* cannot be shown, */
  }                             /* print a warning message */
}  /* error() */

/*--------------------------------------------------------------------*/

static void resize (Dimension size)
{                               /* --- resize widgets */
  Dimension w, h;               /* width and height of top widget */

  size = (Dimension)(size *PTDIST +2 *PTOFFSET +1);
  w    = (Dimension)(mn_minwidth(menu)-2);   /* compute size of board */
  h    = (Dimension)(mn_height(menu) +size); /* and interior widget */
  if (size > w) w = size;       /* get minimal width of board widget */
  XtVaSetValues(w_top,   XtNwidth,    w+2, XtNheight,    h+2,
                         XtNmaxWidth, w+2, XtNmaxHeight, h+2,  NULL);
  XtVaSetValues(w_board, XtNwidth,    w,   XtNheight,    size, NULL);
  XtVaSetValues(w_view,  XtNwidth,    w,   XtNheight,    size, NULL);
  mn_resize(menu, w);           /* set new widget extensions */
}  /* resize() */

/*--------------------------------------------------------------------*/

static void b_draw (int bridge, int hint)
{                               /* --- draw/undraw a bridge */
  unsigned long color;          /* drawing color */
  int x1, y1, x2, y2;           /* coordinates of support points */
  int a;                        /* angle for arc */

  assert(bridge != BI_NONE);    /* check the function arguments */
  if (hint) color = (hint > 0)                     ? green : white;
  else      color = (bi_player(bridge) == BI_HORZ) ? blue  : red;
  XSetForeground(display, gc, color);       /* set the color */
  x1 = bi_col(bridge) *PTDIST/2 +PTOFFSET;  /* get the coordinates */
  y1 = bi_row(bridge) *PTDIST/2 +PTOFFSET;  /* of the crossing point */
  if (bi_dir(bridge) == BI_HORZ)/* draw a horizontal bridge */
    XFillRectangle(display, window, gc, x1-PTDIST/2+2,y1-1, PTDIST-4,3);
  else                          /* draw a vertical   bridge */
    XFillRectangle(display, window, gc, x1-1,y1-PTDIST/2+2, 3,PTDIST-4);
  if (hint >= 0) return;        /* if not to undraw a bridge, abort */
  x2 = x1 -= PTSIZE/2;          /* get upper left corner */
  y2 = y1 -= PTSIZE/2;          /* of crossing point */
  if (bi_dir(bridge) == BI_HORZ) {   /* if the bridge is horizontal, */
    x1 -= PTDIST/2; x2 += PTDIST/2;  /* adapt the horizontal coords. */
    a   = 315*64; }                  /* get right quarter circle */
  else {                             /* if the bridge is vertical, */
    y1 -= PTDIST/2; y2 += PTDIST/2;  /* adapt the vertical coords. */
    a   = 225*64;                    /* get top quarter circle */
  }
  XSetForeground(display, gc,   /* set the player color */
                 (bi_player(bridge) == BI_HORZ) ? blue : red);
  XFillArc(display, window, gc, x1, y1, PTSIZE,PTSIZE, a,        90*64);
  XFillArc(display, window, gc, x2, y2, PTSIZE,PTSIZE, a+180*64, 90*64);
  XSetForeground(display, gc, black);
  XDrawArc(display, window, gc, x1, y1, PTSIZE,PTSIZE, a,        90*64);
  XDrawArc(display, window, gc, x2, y2, PTSIZE,PTSIZE, a+180*64, 90*64);
}  /* b_draw() */               /* redraw the support points */

/*----------------------------------------------------------------------
  Dialog Box Callback Functions
----------------------------------------------------------------------*/

static void dcb_load (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Load Game...' callback */
  FILE    *file;                /* output file */
  BRIDGIT *new;                 /* loaded bridg-it game */
  int     bridge;               /* a bridge built in the game */

  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  file = fopen(fname, "r");     /* open the input file */
  if (!file) { error(E_FOPEN, fname); return; }
  new = bi_load(file);          /* load the bridg-it game */
  fclose(file);                 /* and close the input file */
  if (!new)  { error(E_FREAD, fname); return; }
  if (bi) bi_delete(bi);        /* delete the old bridg-it game */
  bi = new;                     /* and set the new one */
  do {                          /* find player to move next */
    bridge = bi_redo(bi);       /* either by going forward */
    if (bridge != BI_NONE) {    /* and getting the player */
      bi_undo(bi); bip.next = bi_player(bridge);         break; }
    bridge = bi_undo(bi);       /* or by going backwards */
    if (bridge != BI_NONE) {    /* and getting the opponent */
      bi_redo(bi); bip.next = bi_opp(bi_player(bridge)); break; }
    bridge = bip.first;         /* if there is no move, */
  } while (0);                  /* simply get the first player */
  bip.hint = bi_hint(bi, bip.next, drand);
  bip.size = bi_size(bi);       /* get the size of the board */
  resize((Dimension)bip.size);  /* and resize the board */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_load() */

/*--------------------------------------------------------------------*/

static void dcb_save (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Save File...' callback */
  FILE *file;                   /* output file */
  int  r;                       /* result of save function */

  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  file = fopen(fname, "w");     /* open the output file */
  if (!file) { error(E_FOPEN, fname); return; }
  r = bi_save(bi, file);        /* save the bridg-it game */
  if ((fclose(file) != 0) || (r != 0)) {
    error(E_FWRITE, fname); return; }
}  /* dcb_save() */

/*--------------------------------------------------------------------*/

static void dcb_setup (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Setup...' callback */
  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  bip.size = atoi(psh_getval(psh, "size"));
  if (bip.size < MINSIZE) bip.size = MINSIZE;
  if (bip.size > MAXSIZE) bip.size = MAXSIZE;
  bip.first = ((long)psh_getval(psh, "first") == 0) ? BI_HORZ : BI_VERT;
  start(NULL,NULL,NULL,NULL);   /* get the values from the gadgets */
}  /* dcb_setup() */            /* and restart the game */

/*----------------------------------------------------------------------
  Menu Callback Functions
----------------------------------------------------------------------*/

static void mcb_load   (Widget widget, XtPointer client, XtPointer call)
{ load(NULL, NULL, NULL, NULL); }

static void mcb_save   (Widget widget, XtPointer client, XtPointer call)
{ save(NULL, NULL, NULL, NULL); }

static void mcb_quit   (Widget widget, XtPointer client, XtPointer call)
{ quit(NULL, NULL, NULL, NULL); }

static void mcb_start  (Widget widget, XtPointer client, XtPointer call)
{ start(NULL, NULL, NULL, NULL); }

static void mcb_flip   (Widget widget, XtPointer client, XtPointer call)
{ flip(NULL, NULL, NULL, NULL); }

static void mcb_undo   (Widget widget, XtPointer client, XtPointer call)
{ undo(NULL, NULL, NULL, NULL); }

static void mcb_redo   (Widget widget, XtPointer client, XtPointer call)
{ redo(NULL, NULL, NULL, NULL); }

static void mcb_hint   (Widget widget, XtPointer client, XtPointer call)
{ hint(NULL, NULL, NULL, NULL); }

static void mcb_setup  (Widget widget, XtPointer client, XtPointer call)
{ setup(NULL, NULL, NULL, NULL); }

static void mcb_redraw (Widget widget, XtPointer client, XtPointer call)
{ redraw(NULL, NULL, NULL, NULL); }

static void mcb_about  (Widget widget, XtPointer client, XtPointer call)
{ about(NULL, NULL, NULL, NULL); }

/*----------------------------------------------------------------------
  Action Functions
----------------------------------------------------------------------*/

static void mouse (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- mouse button pressed */
  int x, y, px, py;             /* (pixel) coordinates */
  int size;                     /* size of the board */
  int bridge;                   /* bridge to draw */

  assert(bi && gc);             /* check for game and graph. context */
  if (bip.hint == BI_NONE)      /* if the game has already ended, */
    return;                     /* abort the function */
  b_draw(bip.hint, -1);         /* undraw the hint bridge (if any) */

  size = bi_size(bi);           /* get the size of the bridg-it board */
  px = e->xbutton.x -PTOFFSET;  /* get the coordinates at which */
  py = e->xbutton.y -PTOFFSET;  /* the mouse button was pressed */
  x  = px / (PTDIST/2); px %= PTDIST/2;
  y  = py / (PTDIST/2); py %= PTDIST/2;
  if ((x+y) & 1) { if (px > py) x++; else y++; }
  else           { if (px+py > PTDIST/2) { x++; y++; } }
  if ((x <= 0) || (x >= 2*size) /* compute the coordinates of */
  ||  (y <= 0) || (y >= 2*size) /* the closest crossing point */
  ||  (bi_state(bi, x, y) != BI_NONE))
    return;                     /* check for a valid crossing point */

  bridge = bi_bridge(bip.next, x, y);
  bi_build(bi, bridge);         /* construct a bridge, */
  b_draw(bridge, 0);            /* build and draw it */
  XFlush(display);              /* flush buffered drawing commands */
  bip.next = bi_opp(bip.next);  /* switch to the other player */
  bip.hint = bi_hint(bi, bip.next, drand);
  if (bip.hint == BI_NONE) {    /* if there is no new hint */
    db_alert(w_top, dcb_nop, (bip.next == BI_HORZ) ? REDWIN : BLUEWIN);
    return;                     /* notify that the game has ended */
  }                             /* and abort the function */
  bi_build(bi, bip.hint);       /* build the computed bridge */
  b_draw(bip.hint, 0);          /* and draw it */
  bip.next = bi_opp(bip.next);  /* switch to the other player */
  bip.hint = bi_hint(bi, bip.next, drand);
  if (bip.hint == BI_NONE)      /* if there is no new hint */
    db_alert(w_top, dcb_nop, (bip.next == BI_HORZ) ? REDWIN : BLUEWIN);
}  /* mouse() */                /* notify that the game has ended */

/*--------------------------------------------------------------------*/

static void load (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Load... */
  const char *t;                /* buffer for dialog box title */

  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*load"),
    XtNlabel, &t, NULL);        /* get name of menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_load) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* load() */

/*--------------------------------------------------------------------*/

static void save (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Save... */
  const char *t;                /* buffer for dialog box title */

  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*save"),
    XtNlabel, &t, NULL);        /* get name of menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_save) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* save() */

/*--------------------------------------------------------------------*/

static void quit (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Quit */
  if (w_top) XtDestroyApplicationContext(appctx);
  exit(0);                      /* terminate program */
}  /* quit() */

/*--------------------------------------------------------------------*/

static void start (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > New Game */
  BRIDGIT *new;                 /* new bridg-it game */

  assert(bi);                   /* check for an existing game */
  if (bip.size != bi_size(bi)){ /* if the size has been changed, */
    new = bi_create(bip.size);  /* create a new bridg-it game */
    if (!new) { error(E_NOMEM); return; }
    if (bi) bi_delete(bi);      /* delete the old bridg-it game */
    bi = new;                   /* and set the new one */
    resize((Dimension)bip.size);/* resize the window to accomodate */
  }                             /* the new board size */
  bi_init(bi);                  /* initialize the game and */
  bip.next = bip.first;         /* the player to move first */
  bip.hint = bi_hint(bi, bip.next, drand);
  redraw(NULL,NULL,NULL,NULL);  /* get a hint for the first move */
}  /* start() */                /* and redraw the window contents */

/*--------------------------------------------------------------------*/

static void flip (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Flip Sides */
  assert(bi && gc);             /* check for game and graph. context */
  if (bip.hint == BI_NONE)      /* if the game has already ended, */
    return;                     /* abort the function */
  bi_build(bi, bip.hint);       /* build the hint bridge */
  b_draw(bip.hint, 0);          /* and draw it */
  bip.next = bi_opp(bip.next);  /* switch to other player and */
  bip.hint = bi_hint(bi, bip.next, drand);  /* get a new hint */
  if (bip.hint == BI_NONE)      /* if there is no new hint */
    db_alert(w_top, dcb_nop, (bip.next == BI_HORZ) ? REDWIN : BLUEWIN);
}  /* flip() */                 /* notify that the game has ended */

/*--------------------------------------------------------------------*/

static void hint (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Give a Hint */
  assert(bi && gc);             /* check for game and graph. context */
  if (bip.hint != BI_NONE)      /* if there is a hint bridge, */
    b_draw(bip.hint, 1);        /* draw the hint bridge */
}  /* hint() */

/*--------------------------------------------------------------------*/

static void undo (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Undo Move */
  int bridge;                   /* bridge to delete */

  assert(bi && gc);             /* check for game and graph. context */
  if (bip.hint != BI_NONE)      /* if there is a hint for a bridge, */
    b_draw(bip.hint, -1);       /* undraw this bridge */
  bridge = bi_undo(bi);         /* undo a move and get the bridge */
  if (bridge == BI_NONE) return;/* if no move to undo, abort */
  b_draw(bridge, -1);           /* undraw the corresponding bridge */
  bip.next = bi_opp(bip.next);  /* switch to the other player and */
  bip.hint = bi_hint(bi, bip.next, drand);      /* get a new hint */
}  /* undo() */

/*--------------------------------------------------------------------*/

static void redo (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Redo Move */
  int bridge;                   /* bridge to rebuild */

  assert(bi && gc);             /* check for game and graph. context */
  if (bip.hint != BI_NONE)      /* if there is a hint for a bridge, */
    b_draw(bip.hint, -1);       /* undraw this bridge */
  bridge = bi_redo(bi);         /* redo a move and get the bridge */
  if (bridge == BI_NONE) return;/* if there is a move to redo, */
  b_draw(bridge, 0);            /* draw the corresponding bridge */
  bip.next = bi_opp(bip.next);  /* switch to the other player and */
  bip.hint = bi_hint(bi, bip.next, drand);      /* get a new hint */
}  /* redo() */

/*--------------------------------------------------------------------*/

static void setup (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Setup... */
  char buf[32];                 /* buffer for number */

  if (!psh) {                   /* if property sheet not created yet */
    psh = psh_create("setup", w_top, 1);
    if (!psh) return;           /* create a property sheet */
    if ((psh_addline(psh, "lsize")              != 0)
    ||  (psh_additem(psh, "size",  PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh, "lfirst")             != 0)
    ||  (psh_additem(psh, "first", PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh, "first", "blue")      != 0)
    ||  (psh_menuadd(psh, "first", "red")       != 0)) {
      psh_delete(psh); error(E_WIDGET); }
  }
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  sprintf(buf, "%i", bip.size); /* fill the input gadgets */
  psh_setval(psh, "size", buf); /* from the game parameters */
  psh_setval(psh, "first", (char *)((bip.first == BI_HORZ) ? 0L : 1L));
  psh_handle(psh, dcb_setup);   /* handle dialog */
}  /* setup() */

/*--------------------------------------------------------------------*/

static void redraw (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Redraw */
  int        i, n;              /* loop variable, number of moves */
  int        size;              /* size of the board */
  int        x,  y;             /* loop variables, coordinates */
  int        px, py;            /* pixel coordinates */
  Dimension  wd, ht;            /* width and height of window */
  XRectangle rc;                /* clipping rectangle */

  if (!gc) {                    /* if graphics context not created */
    gc = XCreateGC(display, window, 0, NULL);
    if (!gc) { error(E_GRAPH); return; }
  }                             /* create a graphics context */
  if (e && (e->type == Expose)){/* if called by Xt, */
    rc.x      = (short)e->xexpose.x; /* build and set */
    rc.y      = (short)e->xexpose.y; /* clipping rectangle */
    rc.width  = (unsigned short)e->xexpose.width;
    rc.height = (unsigned short)e->xexpose.height;
    XSetClipRectangles(display, gc, 0, 0, &rc, 1, Unsorted);
  }

  /* --- fill background --- */
  XSetForeground(display, gc, white);
  XtVaGetValues(w_board, XtNwidth, &wd, XtNheight, &ht, NULL);
  XFillRectangle(display, window, gc, 0, 0, wd, ht);

  /* --- draw connection panels --- */
  size = bi_size(bi);           /* get the board size and the */
  px   = size *PTDIST;          /* length of the connection panel */
  XSetForeground(display, gc, dred);   /* panels for vert. player */
  XFillRectangle(display, window, gc,
                 PTDIST/2 +PTOFFSET,     PTOFFSET-1,
                 (unsigned)(px -PTDIST), 3);
  XFillRectangle(display, window, gc,
                 PTDIST/2 +PTOFFSET, px +PTOFFSET-1,
                 (unsigned)(px -PTDIST), 3);
  XSetForeground(display, gc, dblue);  /* panels for horz. player */
  XFillRectangle(display, window, gc,
                     PTOFFSET-1, PTDIST/2 +PTOFFSET,
                 3, (unsigned)(px -PTDIST));
  XFillRectangle(display, window, gc,
                 px +PTOFFSET-1, PTDIST/2 +PTOFFSET,
                 3, (unsigned)(px -PTDIST));

  /* --- draw support points --- */
  for (y = size+1; --y >= 0;) { /* traverse the rows */
    py = y *PTDIST +PTOFFSET -PTSIZE/2;
    for (x = size; --x >= 0;) { /* traverse the columns */
      px = x *PTDIST +PTOFFSET +PTDIST/2 -PTSIZE/2;
      XSetForeground(display, gc, red);
      XFillArc(display, window, gc, px, py, PTSIZE, PTSIZE, 0, 360*64);
      XSetForeground(display, gc, black);
      XDrawArc(display, window, gc, px, py, PTSIZE, PTSIZE, 0, 360*64);
    }                           /* draw the support points */
  }                             /* for the red (vertical) player */
  for (y = size; --y >= 0; ) {  /* traverse the rows */
    py = y *PTDIST +PTOFFSET +PTDIST/2 -PTSIZE/2;
    for (x = size+1; --x >= 0;){/* traverse the columns */
      px = x *PTDIST +PTOFFSET -PTSIZE/2;
      XSetForeground(display, gc, blue);
      XFillArc(display, window, gc, px, py, PTSIZE, PTSIZE, 0, 360*64);
      XSetForeground(display, gc, black);
      XDrawArc(display, window, gc, px, py, PTSIZE, PTSIZE, 0, 360*64);
    }                           /* draw the support points */
  }                             /* for the blue (horizontal) player */

  /* --- draw moves made --- */
  for (n = bi_curr(bi), i = 0; i < n; i++)
    b_draw(bi_move(bi, i), 0);  /* draw all bridges built */

  XFlush(display);              /* flush buffered drawing commands */
  XSetClipMask(display, gc, None);      /* and clear clipping mask */
}  /* redraw() */

/*--------------------------------------------------------------------*/

static void about (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- 'About' dialog box */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  if (db_about(w_top, 0, dcb_nop) != 0)
    error(E_DIALOG);            /* show 'about' dialog box */
}  /* about() */

/*----------------------------------------------------------------------
  Main Function
----------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main program */
  Arg       args[8];            /* widget arguments */
  Cardinal  n;                  /* argument counter */
  int       m, w;               /* minimal menu width */
  Colormap  colmap;             /* color map of display */
  XColor    xc[6];              /* buffer for colors */

  rseed((unsigned)time(NULL));  /* init. random number generator */
  XtSetLanguageProc(NULL, (XtLanguageProc)NULL, NULL);

  /* --- create main window --- */
  w_top = XtAppInitialize(&appctx, "Xbi",
            NULL, 0, &argc, argv, fallback, NULL, 0);
  if (!w_top) error(E_INIT);    /* create toplevel widget */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNdefaultDistance, 0);    n++;
  XtSetArg(args[n], XtNresizable,       True); n++;
  w_main = XtCreateManagedWidget("main", formWidgetClass,
             w_top, args, n);   /* create main widget */
  if (!w_main) error(E_WIDGET); /* (form for window layout) */
  menu = mn_create(w_main);     /* create a menu bar and */
  if (!menu) error(E_NOMEM);    /* add titles and items */
  if ((mn_addtitle(menu, "file",     0, 0,          0) != 0)
  ||  (mn_additem (menu, "load",     0, mcb_load,   0) != 0)
  ||  (mn_additem (menu, "save",     0, mcb_save,   0) != 0)
  ||  (mn_additem (menu, NULL,       0, 0,          0) != 0)
  ||  (mn_additem (menu, "quit",     0, mcb_quit,   0) != 0)
  ||  (mn_addtitle(menu, "actions",  0, 0,          0) != 0)
  ||  (mn_additem (menu, "start",    0, mcb_start,  0) != 0)
  ||  (mn_additem (menu, "flip",     0, mcb_flip,   0) != 0)
  ||  (mn_additem (menu, "hint",     0, mcb_hint,   0) != 0)
  ||  (mn_additem (menu, NULL,       0, 0,          0) != 0)
  ||  (mn_additem (menu, "undo",     0, mcb_undo,   0) != 0)
  ||  (mn_additem (menu, "redo",     0, mcb_redo,   0) != 0)
  ||  (mn_additem (menu, NULL,       0, 0,          0) != 0)
  ||  (mn_additem (menu, "setup",    0, mcb_setup,  0) != 0)
  ||  (mn_additem (menu, "redraw",   0, mcb_redraw, 0) != 0)
  ||  (mn_addtitle(menu, "help",     0, 0,          0) != 0)
  ||  (mn_additem (menu, "about",    0, mcb_about,  0) != 0)
  ||  (mn_addtitle(menu, NULL,       0, 0,          0) != 0))
    error(E_WIDGET);

  /* --- create board viewport --- */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNfromVert,  mn_menubar(menu)); n++;
  XtSetArg(args[n], XtNleft,      XtChainLeft);   n++;
  XtSetArg(args[n], XtNtop,       XtChainTop);    n++;
  XtSetArg(args[n], XtNright,     XtChainRight);  n++;
  XtSetArg(args[n], XtNbottom,    XtChainBottom); n++;
  XtSetArg(args[n], XtNresizable, True);          n++;
  w_view = XtCreateManagedWidget("view",
             viewportWidgetClass, w_main, args, n);
  if (!w_view) error(E_WIDGET); /* create viewport */

  /* --- create board widget --- */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 0);             n++;
  XtSetArg(args[n], XtNleft,        XtChainLeft);   n++;
  XtSetArg(args[n], XtNtop,         XtChainTop);    n++;
  XtSetArg(args[n], XtNright,       XtChainRight);  n++;
  XtSetArg(args[n], XtNbottom,      XtChainBottom); n++;
  XtSetArg(args[n], XtNresizable,   True);          n++;
  w_board = XtCreateManagedWidget("board",
              coreWidgetClass, w_view, args, n);
  if (!w_board) error(E_WIDGET);/* create board */

  /* --- set widget extensions --- */
  m = 2*PTDIST +2*PTOFFSET +1;  /* get minimal size of board */
  w = mn_minwidth(menu);        /* and minimal menu width */
  XtVaSetValues(w_top, XtNminWidth,  (w > m) ? w : m,
                       XtNminHeight, mn_height(menu) +m+2, NULL);
  resize((Dimension)bip.size);  /* set board and window size */
  XtOverrideTranslations(w_board, /* install redraw function */
    XtParseTranslationTable("<Expose>: redraw()"));

  /* --- install actions, accelerators and translations --- */
  XtAppAddActions(appctx, actions, XtNumber(actions));
  XtInstallAllAccelerators(w_main, w_main);

  /* --- create bridg-it game --- */
  bi = bi_create(bip.size);     /* create a bridg-game */
  if (!bi) error(E_NOMEM);      /* with a board of default size */
  bi_init(bi);                  /* and initialize the game */
  bip.hint = bi_hint(bi, bip.next, drand);

  /* --- show window --- */
  XtRealizeWidget(w_top);       /* show window on display */
  display = XtDisplay(w_top);   /* get display of top window */
  window  = XtWindow(w_board);  /* and drawable for window */
  if (db_alert(w_top, dcb_nop, NULL) != 0)
    error(E_WIDGET);            /* create alert dialog box */

  /* --- override window manager delete --- */
  XtOverrideTranslations(w_top, /* override window manager delete */
    XtParseTranslationTable("<Message>WM_PROTOCOLS: quit()"));
  wm_delwin = XInternAtom(XtDisplay(w_top), "WM_DELETE_WINDOW", False);
  XSetWMProtocols(XtDisplay(w_top), XtWindow(w_top), &wm_delwin, 1);

  /* --- allocate colors --- */
  screen = DefaultScreen(display);
  colmap = DefaultColormap(display, screen);
  white  = WhitePixel(display, screen);
  black  = BlackPixel(display, screen);
  if ((XAllocNamedColor(display, colmap, "red",   xc+1, xc) == 0)
  ||  (XAllocNamedColor(display, colmap, "blue",  xc+2, xc) == 0)
  ||  (XAllocNamedColor(display, colmap, "green", xc+3, xc) == 0)
  ||  (XAllocNamedColor(display, colmap, "red4",  xc+4, xc) == 0)
  ||  (XAllocNamedColor(display, colmap, "blue4", xc+5, xc) == 0))
    error(E_COLOR);             /* allocate colors and */
  red    = xc[1].pixel;         /* get the pixel values */
  blue   = xc[2].pixel;         /* of the drawing colors */
  green  = xc[3].pixel;
  dred   = xc[4].pixel;
  dblue  = xc[5].pixel;

  /* --- start event loop --- */
  XtAppMainLoop(appctx);        /* enter main event loop */
  return 0;                     /* return 'ok' */
}  /* main() */
